@@ -7,15 +7,30 @@
 """
 import json
 from datetime import datetime
-from Crypto.Signature import PKCS1_v1_5
-from Crypto.Hash import SHA, SHA256
-from Crypto.PublicKey import RSA
 
-from .compat import quote_plus, urlopen, decodebytes, encodebytes, b
+import hashlib
+import OpenSSL
+from Cryptodome.Hash import SHA, SHA256
+from Cryptodome.PublicKey import RSA
+from Cryptodome.Signature import PKCS1_v1_5
+
+from .compat import decodebytes, encodebytes, quote_plus, urlopen
 from .exceptions import AliPayException, AliPayValidationError
 
 
-class BaseAliPay():
+# 常见加密算法
+CryptoAlgSet = (
+    b'rsaEncryption',
+    b'md2WithRSAEncryption',
+    b'md5WithRSAEncryption',
+    b'sha1WithRSAEncryption',
+    b'sha256WithRSAEncryption',
+    b'sha384WithRSAEncryption',
+    b'sha512WithRSAEncryption'
+)
+
+
+class BaseAliPay(object):
     @property
     def appid(self):
         return self._appid
@@ -29,10 +44,6 @@ class BaseAliPay():
         """
         签名用
         """
-        if not self._app_private_key:
-            with open(self._app_private_key_path) as fp:
-                self._app_private_key = RSA.importKey(fp.read())
-
         return self._app_private_key
 
     @property
@@ -40,28 +51,29 @@ class BaseAliPay():
         """
         验证签名用
         """
-        if not self._alipay_public_key:
-            with open(self._alipay_public_key_path) as fp:
-                self._alipay_public_key = RSA.importKey(fp.read())
-
         return self._alipay_public_key
 
-    def __init__(self, appid, app_notify_url, app_private_key_path,
-                 alipay_public_key_path, sign_type="RSA2", debug=False):
+    def __init__(
+        self,
+        appid,
+        app_notify_url,
+        app_private_key_string=None,
+        alipay_public_key_string=None,
+        sign_type="RSA2",
+        debug=False
+    ):
         """
         初始化:
         alipay = AliPay(
           appid="",
-          app_notify_url="",
-          app_private_key_path="",
-          alipay_public_key_path="",
+          app_notify_url="http://example.com",
           sign_type="RSA2"
         )
         """
-        self._appid = appid
+        self._appid = str(appid)
         self._app_notify_url = app_notify_url
-        self._app_private_key_path = app_private_key_path
-        self._alipay_public_key_path = alipay_public_key_path
+        self._app_private_key_string = app_private_key_string
+        self._alipay_public_key_string = alipay_public_key_string
 
         self._app_private_key = None
         self._alipay_public_key = None
@@ -74,16 +86,28 @@ class BaseAliPay():
         else:
             self._gateway = "https://openapi.alipay.com/gateway.do"
 
+        # load key file immediately
+        self._load_key()
+
+    def _load_key(self):
+        # load private key
+        content = self._app_private_key_string
+        self._app_private_key = RSA.importKey(content)
+
+        # load public key
+        content = self._alipay_public_key_string
+        self._alipay_public_key = RSA.importKey(content)
+
     def _sign(self, unsigned_string):
         """
         通过如下方法调试签名
         方法1
-            key = rsa.PrivateKey.load_pkcs1(open(self._app_private_key_path).read())
+            key = rsa.PrivateKey.load_pkcs1(open(self._app_private_key_string).read())
             sign = rsa.sign(unsigned_string.encode("utf8"), key, "SHA-1")
             # base64 编码，转换为unicode表示并移除回车
             sign = base64.encodebytes(sign).decode("utf8").replace("\n", "")
         方法2
-            key = RSA.importKey(open(self._app_private_key_path).read())
+            key = RSA.importKey(open(self._app_private_key_string).read())
             signer = PKCS1_v1_5.new(key)
             signature = signer.sign(SHA.new(unsigned_string.encode("utf8")))
             # base64 编码，转换为unicode表示并移除回车
@@ -96,18 +120,15 @@ class BaseAliPay():
         key = self.app_private_key
         signer = PKCS1_v1_5.new(key)
         if self._sign_type == "RSA":
-            signature = signer.sign(SHA.new(b(unsigned_string)))
+            signature = signer.sign(SHA.new(unsigned_string.encode()))
         else:
-            signature = signer.sign(SHA256.new(b(unsigned_string)))
+            signature = signer.sign(SHA256.new(unsigned_string.encode()))
         # base64 编码，转换为unicode表示并移除回车
         sign = encodebytes(signature).decode("utf8").replace("\n", "")
         return sign
 
     def _ordered_data(self, data):
-        complex_keys = []
-        for key, value in data.items():
-            if isinstance(value, dict):
-                complex_keys.append(key)
+        complex_keys = [k for k, v in data.items() if isinstance(v, dict)]
 
         # 将字典类型的数据dump出来
         for key in complex_keys:
@@ -115,7 +136,9 @@ class BaseAliPay():
 
         return sorted([(k, v) for k, v in data.items()])
 
-    def build_body(self, method, biz_content, return_url=None, append_auth_token=False):
+    def build_body(
+        self, method, biz_content, return_url=None, notify_url=None, append_auth_token=False
+    ):
         data = {
             "app_id": self._appid,
             "method": method,
@@ -130,21 +153,21 @@ class BaseAliPay():
 
         if return_url is not None:
             data["return_url"] = return_url
+
         if method in (
             "alipay.trade.app.pay", "alipay.trade.wap.pay", "alipay.trade.page.pay",
-            "alipay.trade.pay"
-        ):
-            data["notify_url"] = self._app_notify_url
+            "alipay.trade.pay", "alipay.trade.precreate"
+        ) and (notify_url or self._app_notify_url):
+            data["notify_url"] = notify_url or self._app_notify_url
 
         return data
 
     def sign_data(self, data):
         data.pop("sign", None)
         # 排序后的字符串
-        unsigned_items = self._ordered_data(data)
-        unsigned_string = "&".join("{}={}".format(k, v) for k, v in unsigned_items)
-        sign = self._sign(unsigned_string)
         ordered_items = self._ordered_data(data)
+        unsigned_string = "&".join("{}={}".format(k, v) for k, v in ordered_items)
+        sign = self._sign(unsigned_string)
         quoted_string = "&".join("{}={}".format(k, quote_plus(v)) for k, v in ordered_items)
 
         # 获得最终的订单信息字符串
@@ -184,8 +207,10 @@ class BaseAliPay():
             return getattr(self, key)
         raise AttributeError("Unknown attribute" + api_name)
 
-    def api_alipay_trade_wap_pay(self, subject, out_trade_no,
-                                 total_amount, return_url=None, **kwargs):
+    def api_alipay_trade_wap_pay(
+        self, subject, out_trade_no, total_amount,
+        return_url=None, notify_url=None, **kwargs
+    ):
         biz_content = {
             "subject": subject,
             "out_trade_no": out_trade_no,
@@ -193,10 +218,17 @@ class BaseAliPay():
             "product_code": "QUICK_WAP_PAY"
         }
         biz_content.update(kwargs)
-        data = self.build_body("alipay.trade.wap.pay", biz_content, return_url)
+        data = self.build_body(
+            "alipay.trade.wap.pay",
+            biz_content,
+            return_url=return_url,
+            notify_url=notify_url
+        )
         return self.sign_data(data)
 
-    def api_alipay_trade_app_pay(self, subject, out_trade_no, total_amount, **kwargs):
+    def api_alipay_trade_app_pay(
+        self, subject, out_trade_no, total_amount, notify_url=None, **kwargs
+    ):
         biz_content = {
             "subject": subject,
             "out_trade_no": out_trade_no,
@@ -204,11 +236,11 @@ class BaseAliPay():
             "product_code": "QUICK_MSECURITY_PAY"
         }
         biz_content.update(kwargs)
-        data = self.build_body("alipay.trade.app.pay", biz_content)
+        data = self.build_body("alipay.trade.app.pay", biz_content, notify_url=notify_url)
         return self.sign_data(data)
 
     def api_alipay_trade_page_pay(self, subject, out_trade_no, total_amount,
-                                  return_url=None, **kwargs):
+                                  return_url=None, notify_url=None, **kwargs):
         biz_content = {
             "subject": subject,
             "out_trade_no": out_trade_no,
@@ -217,7 +249,12 @@ class BaseAliPay():
         }
 
         biz_content.update(kwargs)
-        data = self.build_body("alipay.trade.page.pay", biz_content, return_url)
+        data = self.build_body(
+            "alipay.trade.page.pay",
+            biz_content,
+            return_url=return_url,
+            notify_url=notify_url
+        )
         return self.sign_data(data)
 
     def api_alipay_trade_query(self, out_trade_no=None, trade_no=None):
@@ -262,7 +299,9 @@ class BaseAliPay():
         raw_string = urlopen(url, timeout=15).read().decode("utf-8")
         return self._verify_and_return_sync_response(raw_string, "alipay_trade_query_response")
 
-    def api_alipay_trade_pay(self, out_trade_no, scene, auth_code, subject, **kwargs):
+    def api_alipay_trade_pay(
+        self, out_trade_no, scene, auth_code, subject, notify_url=None, **kwargs
+    ):
         """
         eg:
             self.api_alipay_trade_pay(
@@ -322,7 +361,7 @@ class BaseAliPay():
             "subject": subject
         }
         biz_content.update(**kwargs)
-        data = self.build_body("alipay.trade.pay", biz_content)
+        data = self.build_body("alipay.trade.pay", biz_content, notify_url=notify_url)
 
         url = self._gateway + "?" + self.sign_data(data)
         raw_string = urlopen(url, timeout=15).read().decode("utf-8")
@@ -371,7 +410,36 @@ class BaseAliPay():
         raw_string = urlopen(url, timeout=15).read().decode("utf-8")
         return self._verify_and_return_sync_response(raw_string, "alipay_trade_cancel_response")
 
-    def api_alipay_trade_precreate(self, subject, out_trade_no, total_amount, **kwargs):
+    def api_alipay_trade_close(self, out_trade_no=None, trade_no=None, operator_id=None):
+        """
+        response = {
+            "alipay_trade_close_response": {
+                "code": "10000",
+                "msg": "Success",
+                "trade_no": "2013112111001004500000675971",
+                "out_trade_no": "YX_001"a
+            }
+        }
+        """
+
+        assert (out_trade_no is not None) or (trade_no is not None),\
+            "Both trade_no and out_trade_no are None"
+
+        biz_content = {}
+        if out_trade_no:
+            biz_content["out_trade_no"] = out_trade_no
+        if trade_no:
+            biz_content["trade_no"] = trade_no
+        if operator_id:
+            biz_content["operator_id"] = operator_id
+
+        data = self.build_body("alipay.trade.close", biz_content)
+
+        url = self._gateway + "?" + self.sign_data(data)
+        raw_string = urlopen(url, timeout=15).read().decode("utf-8")
+        return self._verify_and_return_sync_response(raw_string, "alipay_trade_close_response")
+
+    def api_alipay_trade_precreate(self, subject, out_trade_no, total_amount, notify_url=None, **kwargs):
         """
         success response  = {
           "alipay_trade_precreate_response": {
@@ -400,12 +468,32 @@ class BaseAliPay():
             "subject": subject
         }
         biz_content.update(**kwargs)
-        data = self.build_body("alipay.trade.precreate", biz_content)
+        data = self.build_body("alipay.trade.precreate", biz_content, notify_url=notify_url)
 
         url = self._gateway + "?" + self.sign_data(data)
         raw_string = urlopen(url, timeout=15).read().decode("utf-8")
         return self._verify_and_return_sync_response(raw_string, "alipay_trade_precreate_response")
 
+    def api_alipay_trade_fastpay_refund_query(
+        self, out_request_no, trade_no=None, out_trade_no=None
+    ):
+        assert (out_trade_no is not None) or (trade_no is not None),\
+            "Both trade_no and out_trade_no are None"
+
+        biz_content = {"out_request_no": out_request_no}
+        if trade_no:
+            biz_content["trade_no"] = trade_no
+        else:
+            biz_content["out_trade_no"] = out_trade_no
+
+        data = self.build_body("alipay.trade.fastpay.refund.query", biz_content)
+
+        url = self._gateway + "?" + self.sign_data(data)
+        raw_string = urlopen(url, timeout=15).read().decode("utf-8")
+        return self._verify_and_return_sync_response(
+            raw_string, "alipay_trade_fastpay_refund_query_response"
+        )
+
     def api_alipay_fund_trans_toaccount_transfer(
             self, out_biz_no, payee_type, payee_account, amount, **kwargs
     ):
@@ -443,79 +531,240 @@ class BaseAliPay():
             raw_string, "alipay_fund_trans_order_query_response"
         )
 
+    def api_alipay_trade_order_settle(
+        self,
+        out_request_no,
+        trade_no,
+        royalty_parameters,
+        **kwargs
+    ):
+        biz_content = {
+            "out_request_no": out_request_no,
+            "trade_no": trade_no,
+            "royalty_parameters": royalty_parameters,
+        }
+        biz_content.update(kwargs)
+
+        data = self.build_body("alipay.trade.order.settle", biz_content)
+
+        url = self._gateway + "?" + self.sign_data(data)
+        raw_string = urlopen(url, timeout=15).read().decode("utf-8")
+        return self._verify_and_return_sync_response(
+            raw_string, "alipay_trade_order_settle_response"
+        )
+
     def _verify_and_return_sync_response(self, raw_string, response_type):
         """
-        return data if verification succeeded, else raise exception
+        return response if verification succeeded, raise exception if not
+
+        As to issue #69, json.loads(raw_string)[response_type] should not be returned directly,
+        use json.loads(plain_content) instead
+
+        failed response is like this
+        {
+          "alipay_trade_query_response": {
+            "sub_code": "isv.invalid-app-id",
+            "code": "40002",
+            "sub_msg": "无效的AppID参数",
+            "msg": "Invalid Arguments"
+          }
+        }
         """
 
         response = json.loads(raw_string)
-        result = response[response_type]
+        # raise exceptions
+        if "sign" not in response.keys():
+            result = response[response_type]
+            raise AliPayException(
+                code=result.get("code", "0"),
+                message=raw_string
+            )
+
         sign = response["sign"]
 
         # locate string to be signed
-        raw_string = self._get_string_to_be_signed(
-            raw_string, response_type
-        )
+        plain_content = self._get_string_to_be_signed(raw_string, response_type)
 
-        if not self._verify(raw_string, sign):
+        if not self._verify(plain_content, sign):
             raise AliPayValidationError
-        return result
+        return json.loads(plain_content)
 
     def _get_string_to_be_signed(self, raw_string, response_type):
         """
-        https://doc.open.alipay.com/docs/doc.htm?docType=1&articleId=106120
+        https://docs.open.alipay.com/200/106120
         从同步返回的接口里面找到待签名的字符串
         """
-        left_index = 0
-        right_index = 0
-
-        index = raw_string.find(response_type)
-        left_index = raw_string.find("{", index)
-        index = left_index + 1
-
-        balance = -1
-        while balance < 0 and index < len(raw_string) - 1:
-            index_a = raw_string.find("{", index)
-            index_b = raw_string.find("}", index)
-
-            # 右括号没找到， 退出
-            if index_b == -1:
-                break
-            right_index = index_b
-
-            # 左括号没找到，移动到右括号的位置
-            if index_a == -1:
-                index = index_b + 1
+        balance = 0
+        start = end = raw_string.find("{", raw_string.find(response_type))
+        # 从response_type之后的第一个｛的下一位开始匹配，
+        # 如果是｛则balance加1; 如果是｝而且balance=0，就是待验签字符串的终点
+        for i, c in enumerate(raw_string[start + 1:], start + 1):
+            if c == "{":
                 balance += 1
-            # 左括号出现在有括号之前，移动到左括号的位置
-            elif index_a > index_b:
-                balance += 1
-                index = index_b + 1
-            # 左括号出现在右括号之后， 移动到右括号的位置
-            else:
+            elif c == "}":
+                if balance == 0:
+                    end = i + 1
+                    break
                 balance -= 1
-                index = index_a + 1
-
-        return raw_string[left_index: right_index + 1]
+        return raw_string[start:end]
 
 
 class AliPay(BaseAliPay):
     pass
 
 
+class DCAliPay(BaseAliPay):
+    """
+    数字证书 (digital certificate) 版本
+    """
+    def __init__(
+        self,
+        appid,
+        app_notify_url,
+        app_private_key_string,
+        app_public_key_cert_string,
+        alipay_public_key_cert_string,
+        alipay_root_cert_string,
+        sign_type="RSA2",
+        debug=False
+    ):
+        """
+        初始化
+        DCAlipay(
+            appid='',
+            app_notify_url='http://example.com',
+            app_private_key_string='',
+            app_public_key_cert_string='',
+            alipay_public_key_cert_sring='',
+            aplipay_root_cert_string='',
+        )
+        """
+        self._app_public_key_cert_string = app_public_key_cert_string
+        self._alipay_public_key_cert_string = alipay_public_key_cert_string
+        self._alipay_root_cert_string = alipay_root_cert_string
+        alipay_public_key_string = self.load_alipay_public_key_string()
+        super().__init__(
+            appid=appid,
+            app_notify_url=app_notify_url,
+            app_private_key_string=app_private_key_string,
+            alipay_public_key_string=alipay_public_key_string,
+            sign_type=sign_type,
+            debug=debug
+        )
+
+    def api_alipay_open_app_alipaycert_download(self, alipay_cert_sn):
+        """
+        下载支付宝证书
+        验签使用，支付宝公钥证书无感知升级机制
+        """
+        biz_content = {
+            "alipay_cert_sn": alipay_cert_sn
+        }
+        data = self.build_body("alipay.open.app.alipaycert.download", biz_content)
+        return self.sign_data(data)
+
+    def build_body(self, *args, **kwargs):
+        data = super().build_body(*args, **kwargs)
+        data["app_cert_sn"] = self.app_cert_sn
+        data["alipay_root_cert_sn"] = self.alipay_root_cert_sn
+        return data
+
+    def load_alipay_public_key_string(self):
+        cert = OpenSSL.crypto.load_certificate(
+            OpenSSL.crypto.FILETYPE_PEM, self._alipay_public_key_cert_string
+        )
+        return OpenSSL.crypto.dump_publickey(
+            OpenSSL.crypto.FILETYPE_PEM, cert.get_pubkey()
+        ).decode("utf-8")
+
+    @staticmethod
+    def get_cert_sn(cert):
+        """
+        获取证书 SN 算法
+        """
+        cert = OpenSSL.crypto.load_certificate(OpenSSL.crypto.FILETYPE_PEM, cert)
+        certIssue = cert.get_issuer()
+        name = 'CN={},OU={},O={},C={}'.format(certIssue.CN, certIssue.OU, certIssue.O, certIssue.C)
+        string = name + str(cert.get_serial_number())
+        m = hashlib.md5()
+        m.update(bytes(string, encoding="utf8"))
+        return m.hexdigest()
+
+    @staticmethod
+    def read_pem_cert_chain(certContent):
+        """解析根证书"""
+        certs = list()
+        items = certContent.split('\n\n')  # 根证书中，每个 cert 中间有两个回车间隔
+        items = [i for i in items if i]
+
+        for c in items:
+            cert = OpenSSL.crypto.load_certificate(OpenSSL.crypto.FILETYPE_PEM, c)
+            certs.append(cert)
+        return certs
+
+    @staticmethod
+    def get_root_cert_sn(rootCert):
+        """ 根证书 SN 算法"""
+        certs = DCAliPay.read_pem_cert_chain(rootCert)
+        rootCertSN = None
+        for cert in certs:
+            try:
+                sigAlg = cert.get_signature_algorithm()
+            except ValueError:
+                continue
+            if sigAlg in CryptoAlgSet:
+                certIssue = cert.get_issuer()
+                name = 'CN={},OU={},O={},C={}'.format(
+                    certIssue.CN, certIssue.OU, certIssue.O, certIssue.C
+                )
+                string = name + str(cert.get_serial_number())
+                m = hashlib.md5()
+                m.update(bytes(string, encoding="utf8"))
+                certSN = m.hexdigest()
+                if not rootCertSN:
+                    rootCertSN = certSN
+                else:
+                    rootCertSN = rootCertSN + '_' + certSN
+        return rootCertSN
+
+    @property
+    def app_cert_sn(self):
+        if not hasattr(self, "_app_cert_sn"):
+            self._app_cert_sn = self.get_cert_sn(self._app_public_key_cert_string)
+        return getattr(self, "_app_cert_sn")
+
+    @property
+    def alipay_root_cert_sn(self):
+        if not hasattr(self, "_alipay_root_cert_sn"):
+            self._alipay_root_cert_sn = self.get_root_cert_sn(self._alipay_root_cert_string)
+        return getattr(self, "_alipay_root_cert_sn")
+
+
 class ISVAliPay(BaseAliPay):
 
-    def __init__(self, appid, app_notify_url, app_private_key_path,
-                 alipay_public_key_path, sign_type="RSA2", debug=False,
-                 app_auth_token=None, app_auth_code=None):
+    def __init__(
+        self,
+        appid,
+        app_notify_url,
+        app_private_key_string=None,
+        alipay_public_key_string=None,
+        sign_type="RSA2",
+        debug=False,
+        app_auth_token=None,
+        app_auth_code=None
+    ):
         if not app_auth_token and not app_auth_code:
             raise Exception("Both app_auth_code and app_auth_token are None !!!")
 
         self._app_auth_token = app_auth_token
         self._app_auth_code = app_auth_code
         super(ISVAliPay, self).__init__(
-            appid, app_notify_url, app_private_key_path,
-            alipay_public_key_path, sign_type, debug
+            appid,
+            app_notify_url,
+            app_private_key_string=app_private_key_string,
+            alipay_public_key_string=alipay_public_key_string,
+            sign_type=sign_type,
+            debug=debug
         )
 
     @property
@@ -529,11 +778,14 @@ class ISVAliPay(BaseAliPay):
             raise Exception("Get auth token by auth code failed: {}".format(self._app_auth_code))
         return self._app_auth_token
 
-    def build_body(self, method, biz_content, return_url=None, append_auth_token=True):
+    def build_body(
+        self, method, biz_content, return_url=None, notify_url=None, append_auth_token=True
+    ):
+        return super(ISVAliPay, self).build_body(
+            method, biz_content, return_url, notify_url, append_auth_token
+        )
 
-        return super(ISVAliPay, self).build_body(method, biz_content, return_url, append_auth_token)
-
-    def api_alipay_open_auth_token_app(self, app_auth_code=None, refresh_token=None):
+    def api_alipay_open_auth_token_app(self, refresh_token=None):
         """
         response = {
           "code": "10000",
@@ -547,16 +799,16 @@ class ISVAliPay(BaseAliPay):
         }
         """
 
-        if app_auth_code:
-            biz_content = {
-                "grant_type": "authorization_code",
-                "code": app_auth_code
-            }
-        else:
+        if refresh_token:
             biz_content = {
                 "grant_type": "refresh_token",
                 "refresh_token": refresh_token
             }
+        else:
+            biz_content = {
+                "grant_type": "authorization_code",
+                "code": self._app_auth_code
+            }
         data = self.build_body(
             "alipay.open.auth.token.app",
             biz_content,
